/*p wcnf 4 5 15
15 1 -2 4 0
15 -1 -2 3 0
1 -2 -4 0
1 -3 2 0
1 1 3 0
p wcnf nbvar nbclauses top
*/

#include <fstream>
#include <iostream>
#include <cstring>
#include <cstdio>
#include <sstream>
#include <algorithm>
#include <vector>
#include <cassert>
#include <cstdint>
#include <cmath>
using namespace std;

#include "db.h"

uint64_t numberOfVariables = 0;
uint64_t top = 0;
uint64_t numberOfClauses = 0;
vector< int64_t > values;

uint64_t cost = 0;
uint64_t optimum_found = 0;

bool un_maxsat = false;
string opt = "";

void printFailureAndExit(string s)
{
    cout << "FAIL" << endl;
    cout << s << endl;
    exit(1);
}

class Clause
{
    public:
        Clause(uint64_t weight, bool hard = false)
        {
            hard_ = hard;
            weight_ = weight;
        }

        void addLiteral(int64_t lit)
        {
            literals.push_back(lit);
        }

        void print()
        {
            if(!un_maxsat)
                cout << weight_;
            for( unsigned int i = 0; i < literals.size(); i++ )
                cout << " " << literals[i];
            cout << " 0" << endl;
        }

        string toString()
        {
            stringstream s;
            s << weight_;
            for( unsigned int i = 0; i < literals.size(); i++ )
                s << " " << literals[i];
            s << " 0" << endl;
            return s.str();
        }

        bool check()
        {
            bool satisfied = false;
            for(unsigned int i = 0; i < literals.size(); i++)
            {
                int64_t l = literals[ i ];
                uint64_t al = abs(l);

                if(al >= values.size())
                    printFailureAndExit("Out of bound. Probably the model does not contain all variables.");

                if(values[al] != l)
                    continue;

                satisfied = true;
                break;
            }

            if(hard_ || satisfied)
                return satisfied;

            cost += weight_;
            return true;
        }

    private:
        vector< int64_t > literals;
        uint64_t weight_;
        bool hard_;
};

void readClause(istream& i)
{
    uint64_t weight;
    if(un_maxsat)
        weight=1;
    else
        i >> weight;
    
    if(!i.good())
        printFailureAndExit("Error while reading input file.");

    Clause c(weight, weight == top);
    int64_t tmp;
    while(true)
    {
        i >> tmp;
        if(!i.good())
            printFailureAndExit("Error while reading input file.");
        if(tmp==0)
            break;
        if(tmp>0)
            c.addLiteral(tmp);
        else
            c.addLiteral(tmp);
    }

    if(weight==0)
        return;
    if(!c.check())
        printFailureAndExit(c.toString());
}

void tokenize(const string& str, vector<string>& tokens, const string& delimiters = " ")
{
    // Skip delimiters at beginning.
    string::size_type lastPos = str.find_first_not_of(delimiters, 0);
    // Find first "non-delimiter".
    string::size_type pos     = str.find_first_of(delimiters, lastPos);

    while (string::npos != pos || string::npos != lastPos)
    {
        // Found a token, add it to the vector.
        tokens.push_back(str.substr(lastPos, pos - lastPos));
        // Skip delimiters.  Note the "not_of"
        lastPos = str.find_first_not_of(delimiters, pos);
        // Find next "non-delimiter"
        pos = str.find_first_of(delimiters, lastPos);
    }
}

void parseFirstLine(const string& line)
{
    vector<string> tokens;
    tokenize(line, tokens, " ");
    if(tokens.size() != 4 && tokens.size() != 5)
        printFailureAndExit("Error while reading the file: first line does not match");

    if(tokens[1]=="cnf")
        un_maxsat = true;
    if(!un_maxsat && tokens.size() != 5)
        printFailureAndExit("Error while reading the file: unweighted maxsat instances have no weights");

    uint64_t data[3];
    for(int i = 2; i < tokens.size(); i++)
        data[i-2] = atol(tokens[i].c_str());
    numberOfVariables = data[0];
    numberOfClauses = data[1];

    if(un_maxsat)
        top = 2;
    else
        top = data[2];
}

bool readInput(string filename)
{
    ifstream in;	
	in.open(filename);

    if(!in.is_open())
        printFailureAndExit("Error while opening the file");

    string line;
    while(true)
    {
        if(!in.good())
            printFailureAndExit("Error while reading the file");

        getline(in, line);
        if(line.empty())
            printFailureAndExit("Error while reading the file");

        if(line.at( 0 ) != 'p' )
            continue;
        
        break;        
    }

    for(unsigned int i = 0; i < numberOfClauses; i++)
        readClause(in);

    in.close();
    return true;
}

bool readFirstLine(string filename)
{
    ifstream in;	
	in.open(filename);

    if(!in.is_open())
        printFailureAndExit("Error while opening the file");

    string line;
    while(true)
    {
        if(!in.good())
            printFailureAndExit("Error while reading the file");

        getline(in, line);
        if(line.empty())
            printFailureAndExit("Error while reading the file");

        if(line.at( 0 ) != 'p' )
            continue;
        
        parseFirstLine(line);        
        break;      
    }
    in.close();
    return true;
}

bool compare(int64_t i, int64_t j)
{
    return abs(i) < abs(j);
}

void readModel(string line)
{
    istringstream iss(line);
    string token;
    while(getline(iss, token, ' '))
    {
        if(token == "v")
            continue;
        int64_t value = atoi(token.c_str());
        uint64_t absValue = value < 0 ? -value : value;
        if( absValue > values.size())
            printFailureAndExit("Read a variable that is greater than the original size");
        
        values[absValue] = value;
    }
}

void readOptimum(string line)
{
    istringstream iss(line);
    string token;
    while(getline(iss, token, ' '))
    {
        if(token == "o" || token == "n")
        {
            opt = token;
            continue;
        }
        optimum_found = atoi(token.c_str());
    }
}

bool readModel()
{
    for(unsigned int i = 0; i <= numberOfVariables; i++)
        values.push_back(i);

    string line;
    do
    {
        getline(cin, line);
        if(!cin.good())
            break;
        if(line.empty())
            continue;

        if(line.at(0) == 'v')
        {
            readModel(line);
        }
        else
        {
            if(line.at(0) == 'o' || line.at(0) == 'n')
                readOptimum(line);
            else if(line == "s UNSATISFIABLE")
                printFailureAndExit("Cannot check unsatisfiability");
        }
    } while( true );

    if(values.empty())
        printFailureAndExit("No variables.");
    
//    if( abs(values.back()) != values.size() - 1 )
//        printFailureAndExit("Error while reading the model: some variables were not printed.");

    return true;
}

int main(int argc, char* argv[])
{
    if(argc != 4)
    {
        cerr << "Usage: ./checker exitcode filename istancename" << endl;        
        return 3;
    }

    int exitCode = atoi(argv[1]);
    string filename(argv[2]);
    readFirstLine(filename);
    readModel();
    if( !readInput(filename))
        return 3;

    string instancename(argv[3]);

    if(exitCode == 20)
        printFailureAndExit("Instances are not inconsistent");
    //if(optimum_found != cost)
    //    printFailureAndExit("Optimum found != computed cost");

    DB d;
    uint64_t certifiedOpt = d.instanceToWeight(instancename);
    if(certifiedOpt==-1)
        cerr << "Warning cannot check optimum for file " << filename << endl;
    else// if(opt != "o")
    {    
        if(exitCode == 30)
        {
            if(cost!=certifiedOpt)
                printFailureAndExit("Optimum found does not correspond to certified optimum");
        }
        else if(exitCode == 10)
        {
            if(cost<certifiedOpt)
                printFailureAndExit("Optimal found is less than certified optimum");
        }
        else
        {
            printFailureAndExit("Wrong exit code");        
        }
    }
    /*else
    {
        if(optimum_found<certifiedOpt)
            printFailureAndExit("Suboptimal found claim to be better than certified optimum");
    }*/
    cout << "OK" << endl;
    return 0;
}

